use std::collections::{btree_map::Entry, BTreeMap};

use anyhow::Context;
use itertools::Itertools;
use rayon::prelude::{IntoParallelRefIterator, ParallelIterator};

use crate::sdk::repr::*;
use crate::v2_types::{
    any_type::{self, AnyField, AnyObject, AnyProperty, AnyStruct},
    traits::{
        AsUObject, UArrayPropertyTrait, UBoolPropertyTrait, UEnumPropertyTrait, UEnumTrait,
        UObjectNonConst, UObjectTrait, UPropertyTrait, UStructPropertyTrait, *,
    },
    UClass, UEnum, UFunction, UProperty, UScriptStruct,
};

use super::repr::{
    Class, ClassField, ClassMethod, Enum, Package, ProcessedPackage, Type, UnrealType,
};

fn default_or_anon(name: &str) -> bool {
    name.contains("Default__")
        || name.contains("<uninitialized")
        || name.contains("PLACEHOLDER-CLASS")
}

impl Enum {
    pub fn from_uenum(value: UEnum) -> anyhow::Result<UnrealType> {
        let values = value
            .names()
            .iter()
            .enumerate()
            .map(|(value, name)| {
                let name = name
                    .get_name()
                    .unwrap_or(format!("AnonymousVariant{value}"));
                (value, name)
            })
            .fold(BTreeMap::new(), |mut acc, (value, name)| {
                match acc.entry(name) {
                    Entry::Vacant(entry) => {
                        entry.insert(vec![value]);
                    }
                    Entry::Occupied(mut entry) => {
                        entry.get_mut().push(value);
                    }
                }
                acc
            });

        let values =
            values
                .into_iter()
                .fold(BTreeMap::<u32, String>::new(), |mut acc, (name, values)| {
                    if values.len() > 1 {
                        acc.extend(values.into_iter().map(|i| (i as u32, format!("{name}{i}"))));
                    } else {
                        acc.insert(values.into_iter().next().unwrap() as u32, name);
                    }
                    acc
                });

        Ok(UnrealType::Enum(Enum {
            name: value.get_name().context("could not get name")?,
            full_name: value.get_full_name().context("could not get full name")?,
            values,
        }))
    }
}

impl Class {
    pub fn from_uclass(value: UClass) -> anyhow::Result<UnrealType> {
        todo!()
    }
    pub fn from_uscriptstruct(value: UScriptStruct) -> anyhow::Result<UnrealType> {
        todo!()
    }
}

impl ClassField {
    pub fn from_uprop(prop: UProperty) -> anyhow::Result<Self> {
        Ok(Self {
            offset: *prop.offset() as u32,
            size: *prop.element_size() as u32,
            flags: *prop.property_flags(),
            name: prop
                .name()
                .get_name()
                .context("failed to retrieve field name")?,
            ty: resolve_type(prop).context("failed to get field type")?,
        })
    }
}

impl ClassMethod {
    pub fn from_ufunction(func: UFunction) -> anyhow::Result<Self> {
        let name = func.get_name().context("could not get function name")?;
        let full_name = func
            .get_full_name()
            .context("could not get full function name")?;

        let params = func
            .iter_children()
            .filter_map(|field| match AnyField::from_field(field) {
                AnyField::Property(prop) => Some(prop),
                _ => None,
            })
            .map(|prop| -> anyhow::Result<Option<ClassField>> {
                let param = ClassField {
                    offset: *prop.offset() as u32,
                    size: *prop.element_size() as u32,
                    name: prop.get_name().context("failed to get property name")?,
                    flags: *prop.property_flags(),
                    ty: if let Ok(ty) = resolve_type(prop) {
                        ty
                    } else {
                        log::warn!("skipping field {} with offset 0x{:x}", prop, prop.offset());
                        return Ok(None);
                    },
                };

                Ok(Some(param))
            })
            .fold_ok(Vec::new(), |mut acc, field| {
                acc.extend(field);
                acc
            })?;

        Ok(ClassMethod {
            name,
            full_name,
            flags: *func.function_flags(),
            parameters: params,
        })
    }
}

impl Package {
    pub fn process(self) -> anyhow::Result<ProcessedPackage> {
        self.children
            .par_iter()
            .filter_map(|&object| -> Option<UnrealType> {
                match AnyObject::from_object(object) {
                    AnyObject::Field(field) => match AnyField::from_field(field) {
                        AnyField::Enum(my_enum) => {}
                        AnyField::Struct(my_struct) => match AnyStruct::from_struct(my_struct) {
                            my_struct @ AnyStruct::Class(_)
                            | my_struct @ AnyStruct::ScriptStruct(_) => {}
                            _ => {}
                        },
                        _ => {}
                    },
                    _ => {}
                }

                todo!()
            });
        todo!()
    }
}

fn resolve_type(prop: UProperty) -> anyhow::Result<Type> {
    let ty = match AnyProperty::from_prop(prop.clone()) {
        numeric @ AnyProperty::Numeric(_) => match numeric.as_any_numeric_property().unwrap() {
            any_type::AnyNumericProperty::U8(prop) => match prop.uenum() {
                Some(enm) => Type::Enum {
                    underlying: Box::new(Type::Primitive(PrimitiveType::U8)),
                    enum_type: enm.as_uobject().object_ref(),
                },
                None => Type::Primitive(PrimitiveType::U8),
            },
            any_type::AnyNumericProperty::U16(_) => Type::Primitive(PrimitiveType::U16),
            any_type::AnyNumericProperty::U32(_) => Type::Primitive(PrimitiveType::U32),
            any_type::AnyNumericProperty::U64(_) => Type::Primitive(PrimitiveType::U64),
            any_type::AnyNumericProperty::I8(_) => Type::Primitive(PrimitiveType::I8),
            any_type::AnyNumericProperty::I16(_) => Type::Primitive(PrimitiveType::I16),
            any_type::AnyNumericProperty::I32(_) => Type::Primitive(PrimitiveType::I32),
            any_type::AnyNumericProperty::I64(_) => Type::Primitive(PrimitiveType::I64),
            any_type::AnyNumericProperty::F32(_) => Type::Primitive(PrimitiveType::F32),
            any_type::AnyNumericProperty::F64(_) => Type::Primitive(PrimitiveType::F64),
            any_type::AnyNumericProperty::Other(_) => {
                return Err(anyhow::anyhow!("unhandled numeric property"));
            }
        },
        AnyProperty::Bool(prop) => Type::Primitive(PrimitiveType::Bool {
            byte_mask: *prop.byte_mask(),
            field_mask: *prop.field_mask(),
            byte_offset: *prop.byte_offset(),
            field_size: *prop.field_size(),
        }),
        AnyProperty::Interface(_) => return Err(anyhow::anyhow!("skipping interfaces for now")),
        object @ AnyProperty::ObjectBase(_) => {
            match object.as_any_object_base_property().unwrap() {
                any_type::AnyObjectBaseProperty::Object(obj) => Type::Class(
                    obj.property_class()
                        .context("object property missing properry class")?
                        .as_uobject()
                        .object_ref(),
                ),
                any_type::AnyObjectBaseProperty::WeakObject(obj) => Type::WeakPtr(
                    obj.property_class()
                        .context("weak ptr property missing property class.")?
                        .as_uobject()
                        .object_ref(),
                ),
                any_type::AnyObjectBaseProperty::LazyObject(obj) => Type::LazyPtr(
                    obj.property_class()
                        .context("lazy ptr property missing property class.")?
                        .as_uobject()
                        .object_ref(),
                ),
                any_type::AnyObjectBaseProperty::SoftObject(obj) => Type::SoftPtr(
                    obj.property_class()
                        .context("soft ptr property missing property class.")?
                        .as_uobject()
                        .object_ref(),
                ),
                asset @ any_type::AnyObjectBaseProperty::AssetObject(_) => {
                    match asset.as_any_asset_object_property().unwrap() {
                        any_type::AnyAssetObjectProperty::Class(class) => Type::AssetPtr(
                            class
                                .property_class()
                                .context("asset object property missing properry class")?
                                .as_uobject()
                                .object_ref(),
                        ),
                        any_type::AnyAssetObjectProperty::Object(_) => {
                            return Err(anyhow::anyhow!(
                                "unhandled asset object property (NOT AN ERROR)"
                            ));
                        }
                    }
                }
                any_type::AnyObjectBaseProperty::Other(_) => {
                    return Err(anyhow::anyhow!("unhandled object base property"));
                }
            }
        }
        AnyProperty::Array(array) => Type::Array(Box::new(resolve_type(
            array
                .inner()
                .context("array property inner type missing.")?,
        )?)),
        AnyProperty::Map(_) => unreachable!("not used in ARK"),
        AnyProperty::Str(_) => Type::String,
        AnyProperty::Text(_) => Type::Text,
        AnyProperty::Name(_) => Type::Name,
        AnyProperty::Delegate(_) => {
            return Err(anyhow::anyhow!("skipping delegates for now"));
        }
        AnyProperty::MulticastDelegate(_) => {
            return Err(anyhow::anyhow!("skipping multicast delegates for now"));
        }
        AnyProperty::Enum(enm) => Type::Enum {
            underlying: Box::new(resolve_type(
                enm.underlying_type()
                    .context("enum property was missing underlying type")?,
            )?),
            enum_type: enm
                .uenum()
                .context("enum property missing enum type")?
                .as_uobject()
                .object_ref(),
        },
        AnyProperty::Struct(class) => Type::Struct(
            class
                .ustruct()
                .context("struct property had no inner struct")?
                .as_uobject()
                .object_ref(),
        ),
        AnyProperty::Other(_) => {
            return Err(anyhow::anyhow!("unhandled property."));
        }
    };

    if *prop.array_dim() > 1 {
        Ok(Type::RawArray {
            ty: Box::new(ty),
            len: *prop.array_dim() as u32,
        })
    } else {
        Ok(ty)
    }
}
